#!/usr/bin/python3

stock_quality = "MP3_320"

qualities = {
	"MP3_320": {
		"n_quality": "3",
		"f_format": ".mp3",
		"s_quality": "320"
	},

	"FLAC": {
		"n_quality": "9",
		"f_format": ".flac",
		"s_quality": "FLAC"
	},

	"MP3_128": {
		"n_quality": "1",
		"f_format": ".mp3",
		"s_quality": "128"
	}
}